@@ -47,4 +47,4 @@
 
   (add-type-alias! ws "AtkState" 'long-long)
   
-  (load-defs ws "gnome/defs/atk.defs"))
+  (load-defs-with-overrides ws "gnome/defs/atk.defs"))
