@@ -33,11 +33,25 @@
 
 
 
+
+typedef struct {
+    SCM (*wrap) (const GValue*);
+    void (*unwrap) (SCM, GValue*);
+} wrap_funcs;
+
+typedef struct {
+    SCMGValueGetTypeInstanceFunc getter;
+    SCMGValueSetTypeInstanceFunc setter;
+} gtype_instance_wrap_funcs;
+
 scm_t_bits scm_tc16_gvalue;
 
 SCM_SYMBOL (sym_closure, "closure");
 SCM_KEYWORD (k_value, "value");
 
+static GHashTable *gvalue_wrappers = NULL;
+static GHashTable *gtype_instance_wrappers = NULL;
+
 static SCM _make;
 static SCM _gtype_instance_write;
 
@@ -198,10 +212,14 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
 #define FUNC_NAME s_scm_gvalue_primitive_set
 {
     GValue *gvalue;
+    GType fundamental;
+    gtype_instance_wrap_funcs *w;
 
     SCM_VALIDATE_GVALUE_COPY (1, instance, gvalue);
 
-    switch (G_TYPE_FUNDAMENTAL (G_VALUE_TYPE (gvalue))) {
+    fundamental = G_TYPE_FUNDAMENTAL (G_VALUE_TYPE (gvalue));
+    
+    switch (fundamental) {
     case G_TYPE_CHAR:
 	SCM_VALIDATE_CHAR (2, value);
 	g_value_set_char (gvalue, SCM_CHAR (value));
@@ -288,36 +306,6 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
         break;
     }
 
-    /* could probably add G_TYPE_PARAM as well */
-    case G_TYPE_OBJECT:
-    case G_TYPE_INTERFACE: /* assuming GObject is a prereq */ {
-        if (SCM_FALSEP (value)) {
-            g_value_set_object (gvalue, NULL);
-        } else {
-            gpointer ginstance;
-            GType gtype;
-
-            if (SCM_GOBJECTP (value)) {
-                SCM_VALIDATE_GOBJECT_COPY (2, value, ginstance);
-            }
-            else {
-                SCM_VALIDATE_GTYPE_INSTANCE_COPY (2, value, ginstance);
-            }
-            gtype = G_TYPE_FROM_INSTANCE (ginstance);
-
-            switch (G_TYPE_FUNDAMENTAL (gtype)) {
-            case G_TYPE_OBJECT:
-                SCM_ASSERT (g_type_is_a (gtype, G_VALUE_TYPE (gvalue)),
-                            value, SCM_ARG2, FUNC_NAME);
-                g_value_set_object (gvalue, ginstance);
-                break;
-            default:
-                SCM_ERROR_NOT_YET_IMPLEMENTED (value);
-            }
-        }
-        break;
-    }
-
     case G_TYPE_BOXED:
         if (G_VALUE_TYPE (gvalue) == G_TYPE_BOXED_SCM) {
             /* the copy func will protect the scm_value from GC */
@@ -369,7 +357,30 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
         break;
 
     default:
-	scm_wrong_type_arg (FUNC_NAME, SCM_ARG1, instance);
+        if ((w = g_hash_table_lookup (gtype_instance_wrappers,
+                                      GINT_TO_POINTER (fundamental)))) {
+            if (SCM_FALSEP (value)) {
+                w->setter (gvalue, NULL);
+            } else {
+                gpointer ginstance;
+                GType gtype;
+
+                if (SCM_GOBJECTP (value)) {
+                    SCM_VALIDATE_GOBJECT_COPY (2, value, ginstance);
+                }
+                else {
+                    SCM_VALIDATE_GTYPE_INSTANCE_COPY (2, value, ginstance);
+                }
+                gtype = G_TYPE_FROM_INSTANCE (ginstance);
+
+                SCM_ASSERT (g_type_is_a (gtype, G_VALUE_TYPE (gvalue)),
+                            value, SCM_ARG2, FUNC_NAME);
+
+                w->setter (gvalue, ginstance);
+            }
+        } else {
+            scm_wrong_type_arg (FUNC_NAME, SCM_ARG1, instance);
+        }
 	break;
     }
 
@@ -458,6 +469,7 @@ SCM_DEFINE (scm_gvalue_primitive_get, "gvalue-primitive-get", 1, 0, 0,
         break;
     }
 
+        /* FIXME: why are these here? */
     case G_TYPE_OBJECT:
         /* scm_c_gtype_instance_to_scm will ref the object for us */
 	return scm_c_gtype_instance_to_scm (g_value_get_object (gvalue));
@@ -501,12 +513,6 @@ scm_c_dup_gboxed_to_scm (GType boxed_type, gconstpointer boxed_value)
     return svalue;
 }
 
-static GHashTable *gvalue_wrappers = NULL;
-typedef struct {
-    SCM (*wrap) (const GValue*);
-    void (*unwrap) (SCM, GValue*);
-} wrap_funcs;
-
 /* not threadsafe */
 void
 scm_c_register_gvalue_wrappers (GType type,
@@ -524,6 +530,21 @@ scm_c_register_gvalue_wrappers (GType type,
     g_hash_table_insert (gvalue_wrappers, GINT_TO_POINTER (type), w);
 }
 
+void
+scm_c_register_gtype_instance_gvalue_wrappers (GType type,
+    SCMGValueGetTypeInstanceFunc getter, SCMGValueSetTypeInstanceFunc setter)
+{
+    gtype_instance_wrap_funcs* w = g_new (gtype_instance_wrap_funcs, 1);
+
+    if (!gtype_instance_wrappers)
+        gtype_instance_wrappers = g_hash_table_new (NULL, NULL);
+    
+    w->getter = getter;
+    w->setter = setter;
+    
+    g_hash_table_insert (gtype_instance_wrappers, GINT_TO_POINTER (type), w);
+}
+
 SCM_DEFINE (scm_gvalue_to_scm, "gvalue->scm", 1, 0, 0,
 	    (SCM value),
 	    "")
@@ -533,6 +554,7 @@ SCM_DEFINE (scm_gvalue_to_scm, "gvalue->scm", 1, 0, 0,
 
     GType type, fundamental;
     SCM stype;
+    gtype_instance_wrap_funcs* w;
 
     SCM_VALIDATE_GVALUE_COPY (1, value, gvalue);
 
@@ -544,11 +566,9 @@ SCM_DEFINE (scm_gvalue_to_scm, "gvalue->scm", 1, 0, 0,
     if (SCM_NFALSEP (scm_gtype_basic_p (stype))) {
         return scm_gvalue_primitive_get (value);
     } 
-    else if (fundamental == G_TYPE_OBJECT) {
-        return scm_c_gtype_instance_to_scm (g_value_get_object (gvalue));
-    }
-    else if (fundamental == G_TYPE_PARAM) {
-        return scm_c_gtype_instance_to_scm (g_value_get_param (gvalue));
+    else if ((w = g_hash_table_lookup (gtype_instance_wrappers,
+                                       GINT_TO_POINTER (fundamental)))) {
+        return scm_c_gtype_instance_to_scm (w->getter (gvalue));
     }
     else if (gvalue_wrappers) {
         wrap_funcs* w;
@@ -572,13 +592,12 @@ SCM_DEFINE (scm_gvalue_to_scm, "gvalue->scm", 1, 0, 0,
 SCM scm_c_gvalue_to_scm (const GValue *gvalue)
 {
     GType fundamental = G_TYPE_FUNDAMENTAL (G_VALUE_TYPE (gvalue));
+    gtype_instance_wrap_funcs* w;
     
     /* try to avoid needless allocation of a gvalue smob */
-    if (fundamental == G_TYPE_OBJECT) {
-        return scm_c_gtype_instance_to_scm (g_value_get_object (gvalue));
-    }
-    else if (fundamental == G_TYPE_PARAM) {
-        return scm_c_gtype_instance_to_scm (g_value_get_param (gvalue));
+    if ((w = g_hash_table_lookup (gtype_instance_wrappers,
+                                  GINT_TO_POINTER (fundamental)))) {
+        return scm_c_gtype_instance_to_scm (w->getter (gvalue));
     }
     else {
         /* fall back on the normal version */
@@ -615,7 +634,8 @@ SCM_DEFINE (scm_scm_to_gvalue, "scm->gvalue", 2, 0, 0,
         /* Some types can be converted to GValues with the same incantation. */
         return scm_call_3 (_make, scm_c_gtype_to_class (gtype), k_value, scm);
     }
-    else if (fundamental == G_TYPE_OBJECT || fundamental == G_TYPE_INTERFACE) {
+    else if (g_hash_table_lookup (gtype_instance_wrappers,
+                                  GINT_TO_POINTER (fundamental))) {
         SCM ret = scm_gvalue_primitive_new (type);
         scm_gvalue_primitive_set (ret, (SCM_FALSEP (scm) 
                                         ? scm
